from typing import Any

import awswrangler as wr
import numpy as np
import pandas as pd
import toolz as tlz

from ..base import Cell, IncrementalCell
from ..triangle import Triangle

__all__ = [
    "triangle_to_wide_csv",
    "triangle_to_long_csv",
    "triangle_to_wide_data_frame",
    "triangle_to_long_data_frame",
]


def triangle_to_wide_csv(tri: Triangle[Cell], filename: str) -> None:
    """Convert a Triangle to a wide-format CSV file.

    See `wide_data_frame_to_triangle` for details on the wide format.

    Args:
        tri: Triangle to convert.
        filename: Name of CSV file to create.
    """
    df = triangle_to_wide_data_frame(tri)
    # noinspection PyTypeChecker
    if filename.startswith("s3://"):
        wr.s3.to_csv(df, filename, index=False)
    else:
        df.to_csv(filename, index=False)


def triangle_to_long_csv(tri: Triangle[Cell], filename: str) -> None:
    """Convert a Triangle to a long-format CSV file.

    See `long_data_frame_to_triangle` for details on the long format.

    Args:
        tri: Triangle to convert.
        filename: Name of CSV file to create.
    """
    df = triangle_to_long_data_frame(tri)
    # noinspection PyTypeChecker
    if filename.startswith("s3://"):
        wr.s3.to_csv(df, filename, index=False)
    else:
        df.to_csv(filename, index=False)


def triangle_to_wide_data_frame(tri: Triangle[Cell]) -> pd.DataFrame:
    """Convert a Triangle to a wide-format Pandas `DataFrame`.

    See `wide_data_frame_to_triangle` for details on the wide format.

    Args:
        tri: Triangle to convert.
    """
    metadata_names = _all_metadata_names(tri)
    field_names = _all_fields(tri)
    rows = [
        row
        for cell in tri
        for row in _cell_to_wide_dict(cell, metadata_names, field_names)
    ]

    df = pd.DataFrame(rows)
    df = _drop_constant_scenario(df)
    df.period_start = pd.to_datetime(df.period_start)
    df.period_end = pd.to_datetime(df.period_end)
    df.evaluation_date = pd.PeriodIndex(
        _date_to_period_index(df.evaluation_date)
    ).to_timestamp()
    if tri.is_incremental:
        df.prev_evaluation_date = pd.PeriodIndex(
            _date_to_period_index(df.prev_evaluation_date)
        )
    return df


def triangle_to_long_data_frame(tri: Triangle[Cell]) -> pd.DataFrame:
    """Convert a Triangle to a long-format Pandas `DataFrame`.

    See `long_data_frame_to_triangle` for details on the long format.

    Args:
        tri: Triangle to convert.
    """
    metadata_names = _all_metadata_names(tri)
    rows = [row for cell in tri for row in _cell_to_long_dict(cell, metadata_names)]

    df = pd.DataFrame(rows)
    df = _drop_constant_scenario(df)
    df.period_start = pd.to_datetime(df.period_start)
    df.period_end = pd.to_datetime(df.period_end)
    df.evaluation_date = pd.PeriodIndex(_date_to_period_index(df.evaluation_date))
    if tri.is_incremental:
        df.prev_evaluation_date = pd.PeriodIndex(
            _date_to_period_index(df.prev_evaluation_date)
        )
    return df


def _cell_to_long_dict(cell: Cell, metadata_names: list[str]) -> list[dict[str, Any]]:
    metadata_raw_dict = cell.metadata.as_flat_dict()
    metadata_dict = {name: metadata_raw_dict.get(name) for name in metadata_names}
    base_dict = {
        "period_start": cell.period_start,
        "period_end": cell.period_end,
        "evaluation_date": cell.evaluation_date,
    }
    if isinstance(cell, IncrementalCell):
        base_dict.update({"prev_evaluation_date": cell.prev_evaluation_date})
    clean_field_dicts = _clean_field_dicts(cell)

    constant_fields = tlz.valfilter(
        lambda vals: isinstance(vals, (float, int)), cell.values
    )

    return [
        {
            **base_dict,
            **metadata_dict,
            "scenario": np.NaN if field in constant_fields else ndx + 1,
            "field": field,
            "value": float(value),
        }
        for ndx, field_dict in enumerate(clean_field_dicts)
        for field, value in field_dict.items()
    ]


def _cell_to_wide_dict(
    cell: Cell, metadata_names: list[str], field_names: list[str]
) -> list[dict[str, Any]]:
    metadata_raw_dict = cell.metadata.as_flat_dict()
    metadata_dict = {name: metadata_raw_dict.get(name) for name in metadata_names}
    clean_field_dicts = _clean_field_dicts(cell, field_names)
    base_dict = {
        "period_start": cell.period_start,
        "period_end": cell.period_end,
        "evaluation_date": cell.evaluation_date,
    }
    if isinstance(cell, IncrementalCell):
        base_dict.update({"prev_evaluation_date": cell.prev_evaluation_date})

    return [
        {
            **base_dict,
            "scenario": ndx + 1,
            **field_dict,
            **metadata_dict,
        }
        for ndx, field_dict in enumerate(clean_field_dicts)
    ]


def _clean_field_dicts(
    cell: Cell, field_names: list[str] = None
) -> list[dict[str, float]]:
    if field_names is None:
        field_names = list(cell.values.keys())
    common_length = _common_field_length(cell, field_names)
    field_dicts = []

    for ndx in range(common_length):
        field_dict = {}
        for field in field_names:
            value = cell.values.get(field)
            if isinstance(value, np.ndarray) and value.size > 1:
                field_dict[field] = value[ndx]
            elif value is not None:
                field_dict[field] = float(value)
        field_dicts.append(field_dict)

    return field_dicts


def _common_field_length(cell: Cell, field_names: list[str]) -> int:
    field_lengths = set()

    for field in field_names:
        value = cell.values.get(field)
        if isinstance(value, np.ndarray):
            if len(value.shape) > 1:
                raise ValueError("Multidimensional array values are not supported yet")
            field_lengths |= {value.size}
        else:
            field_lengths |= {1}

    # We don't want multiple non-1 field_lengths
    if field_lengths == {1}:
        return 1
    interesting_field_lengths = field_lengths - {1}
    if len(interesting_field_lengths) > 1:
        raise ValueError("Incompatible field lengths are not supported yet")
    return max(interesting_field_lengths)


def _all_fields(tri: Triangle[Cell]) -> list[str]:
    field_names = set()

    for cell in tri:
        cell_fields = set(cell.values.keys())
        field_names |= cell_fields

    return list(field_names)


def _all_metadata_names(tri: Triangle[Cell]) -> list[str]:
    metadata_col_names = set()
    all_metadata = tri.metadata

    for metadata in all_metadata:
        col_names = {k for k, v in metadata.as_flat_dict().items() if v is not None}
        metadata_col_names |= col_names

    return list(metadata_col_names)


def _date_to_period_index(dates: list) -> list[pd.Period]:
    return [pd.Period(year=d.year, month=d.month, day=d.day, freq="D") for d in dates]


def _drop_constant_scenario(df: pd.DataFrame) -> pd.DataFrame:
    scenario_values = df["scenario"].to_numpy()
    if np.all(scenario_values[0] == scenario_values) or np.all(
        pd.isnull(scenario_values)
    ):
        return df.drop(columns=["scenario"])
    else:
        return df



from .adjust import *
from .aggregate import *
from .backfill import *
from .basis import accident_quarter_to_policy_year, to_cumulative, to_incremental
from .bootstrap import bootstrap
from .currency import *
from .disaggregate import *
from .extend import *
from .fields import *
from .fill import *
from .join import *
from .merge import *
from .method_moments import *
from .shift_origin import *
from .slice import *
from .summarize import *
from .thin import *



from functools import wraps

from .io import (
    array_data_frame_to_triangle,
    binary_to_triangle,
    chain_ladder_to_triangle,
    json_to_triangle,
    long_csv_to_triangle,
    long_data_frame_to_triangle,
    statics_data_frame_to_triangle,
    triangle_to_array_data_frame,
    triangle_to_binary,
    triangle_to_chain_ladder,
    triangle_to_json,
    triangle_to_long_csv,
    triangle_to_long_data_frame,
    triangle_to_right_edge_data_frame,
    triangle_to_wide_csv,
    triangle_to_wide_data_frame,
    wide_csv_to_triangle,
    wide_data_frame_to_triangle,
)
from .triangle import Triangle
from .utils import (
    add_statics,
    aggregate,
    blend,
    coalesce,
    make_right_diagonal,
    make_right_triangle,
    merge,
    period_merge,
    split,
    summarize,
    thin,
    to_cumulative,
    to_incremental,
)

# utils
Triangle.aggregate = wraps(aggregate)(
    lambda self, *args, **kwargs: aggregate(self, *args, **kwargs)
)
Triangle.summarize = wraps(summarize)(
    lambda self, *args, **kwargs: summarize(self, *args, **kwargs)
)
Triangle.blend = wraps(blend)(
    lambda self, triangles, **kwargs: blend(triangles=[self, *triangles], **kwargs)
)
Triangle.split = wraps(split)(
    lambda self, *args, **kwargs: split(self, *args, **kwargs)
)
Triangle.merge = wraps(merge)(
    lambda self, *args, **kwargs: merge(self, *args, **kwargs)
)
Triangle.period_merge = wraps(period_merge)(
    lambda self, *args, **kwargs: period_merge(self, *args, **kwargs)
)
Triangle.coalesce = wraps(coalesce)(
    lambda self, triangles: coalesce([self, *triangles])
)
Triangle.to_incremental = wraps(to_incremental)(lambda self: to_incremental(self))
Triangle.to_cumulative = wraps(to_cumulative)(lambda self: to_cumulative(self))
Triangle.add_statics = wraps(add_statics)(
    lambda self, *args, **kwargs: add_statics(self, *args, **kwargs)
)
Triangle.make_right_triangle = wraps(make_right_triangle)(
    lambda self, *args, **kwargs: make_right_triangle(self, *args, **kwargs)
)
Triangle.make_right_diagonal = wraps(make_right_diagonal)(
    lambda self, *args, **kwargs: make_right_diagonal(self, *args, **kwargs)
)
Triangle.thin = wraps(thin)(lambda self, *args, **kwargs: thin(self, *args, **kwargs))

# io
Triangle.to_array_data_frame = wraps(triangle_to_array_data_frame)(
    lambda self, *args, **kwargs: triangle_to_array_data_frame(self, *args, **kwargs)
)
Triangle.to_binary = wraps(triangle_to_binary)(
    lambda self, *args, **kwargs: triangle_to_binary(self, *args, **kwargs)
)
Triangle.to_json = wraps(triangle_to_json)(triangle_to_json)
Triangle.to_chain_ladder = wraps(triangle_to_chain_ladder)(
    lambda self, *args, **kwargs: triangle_to_chain_ladder(self, *args, **kwargs)
)
Triangle.to_long_csv = wraps(triangle_to_long_csv)(
    lambda self, *args, **kwargs: triangle_to_long_csv(self, *args, **kwargs)
)
Triangle.to_long_data_frame = wraps(triangle_to_long_data_frame)(
    lambda self, *args, **kwargs: triangle_to_long_data_frame(self, *args, **kwargs)
)
Triangle.to_right_edge_data_frame = wraps(triangle_to_right_edge_data_frame)(
    lambda self, *args, **kwargs: triangle_to_right_edge_data_frame(
        self, *args, **kwargs
    )
)
Triangle.to_wide_csv = wraps(triangle_to_wide_csv)(
    lambda self, *args, **kwargs: triangle_to_wide_csv(self, *args, **kwargs)
)
Triangle.to_wide_data_frame = wraps(triangle_to_wide_data_frame)(
    lambda self, *args, **kwargs: triangle_to_wide_data_frame(self, *args, **kwargs)
)

# constructors
Triangle.from_array_data_frame = staticmethod(
    wraps(array_data_frame_to_triangle)(array_data_frame_to_triangle)
)
Triangle.from_binary = staticmethod(wraps(binary_to_triangle)(binary_to_triangle))
Triangle.from_chain_ladder = staticmethod(
    wraps(chain_ladder_to_triangle)(chain_ladder_to_triangle)
)
Triangle.from_long_csv = staticmethod(wraps(long_csv_to_triangle)(long_csv_to_triangle))
Triangle.from_long_data_frame = staticmethod(
    wraps(long_data_frame_to_triangle)(long_data_frame_to_triangle)
)
Triangle.from_statics_data_frame = staticmethod(
    wraps(statics_data_frame_to_triangle)(statics_data_frame_to_triangle)
)
Triangle.from_wide_csv = staticmethod(wraps(wide_csv_to_triangle)(wide_csv_to_triangle))
Triangle.from_wide_data_frame = staticmethod(
    wraps(wide_data_frame_to_triangle)(wide_data_frame_to_triangle)
)
Triangle.from_json = staticmethod(wraps(json_to_triangle)(json_to_triangle))



import datetime
import json
import warnings
from typing import IO, Any, Union

import numpy as np

from ..base import Cell, CumulativeCell, IncrementalCell, Metadata
from ..triangle import Triangle

__all__ = [
    "json_to_triangle",
    "json_string_to_triangle",
    "triangle_json_loads",
    "triangle_json_load",
    "triangle_to_json",
]


def json_to_triangle(
    file_or_fname: Union[str, IO], date_format: str = "%Y-%m-%d"
) -> Triangle:
    if isinstance(file_or_fname, str):
        with open(file_or_fname, "r") as infile:
            cells = json.load(infile, cls=TriangleDecoder, date_format=date_format)
    else:
        cells = json.load(file_or_fname, cls=TriangleDecoder, date_format=date_format)
    return Triangle(cells)


def json_string_to_triangle(string: str, date_format: str = "%Y-%m-%d") -> Triangle:
    cells = json.loads(string, cls=TriangleDecoder, date_format=date_format)
    return Triangle(cells)


def triangle_json_loads(string: str, date_format: str = "%Y-%m-%d") -> Triangle:
    warnings.warn(
        "triangle_json_loads is being deprecated, use json_string_to_triangle instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return json_string_to_triangle(string, date_format)


def triangle_json_load(file: IO, date_format: str = "%Y-%m-%d") -> Triangle:
    warnings.warn(
        "triangle_json_load is being deprecated, use json_to_triangle instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return json_to_triangle(file, date_format)


def triangle_to_json(
    tri: Triangle[Cell], file_or_fname: Union[str, IO] | None = None
) -> str | None:
    """Convert a triangle to JSON.

    Args:
        tri: A triangle to convert to a JSON representation.
        file_or_fname: A filename, a file handle, or `None`.
    Returns:
        If `file_or_fname` is `None`, then a string with the JSON representation of the
        triangle is returned. Otherwise, the JSON representation is written to the provided
        file and `None` is returned.
    """
    if file_or_fname:
        # The file_or_fname behavior is the same as triangle_to_wide_csv and others.
        if isinstance(file_or_fname, str):
            with open(file_or_fname, "w") as outfile:
                json.dump(tri, outfile, cls=TriangleEncoder)
        else:
            json.dump(tri, file_or_fname, cls=TriangleEncoder)
    else:
        # The return-a-string behavior is to maintain backwards compatability with older
        # versions of Bermuda that didn't support directly writing JSON to a file.
        return json.dumps(tri, cls=TriangleEncoder)


class TriangleDecoder(json.JSONDecoder):
    def __init__(self, date_format="%Y-%m-%d", *args, **kwargs):
        self.date_format = date_format
        super().__init__(object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if "slices" in obj:
            return sum(obj["slices"], [])
        if "cells" in obj:
            return self._parse_cell_set(obj)
        if "period_start" in obj and "period_end" in obj and "values" in obj:
            return self._parse_observation(obj)
        return obj

    @staticmethod
    def _parse_cell_set(obj):
        metadata = Metadata(
            risk_basis=obj.get("risk_basis", "Accident"),
            country=obj.get("country"),
            currency=obj.get("currency"),
            reinsurance_basis=obj.get("reinsurance_basis"),
            loss_definition=obj.get("loss_definition"),
            per_occurrence_limit=obj.get("per_occurrence_limit"),
            details=obj.get("details", {}),
            loss_details=obj.get("loss_details", {}),
        )

        return [ob.replace(metadata=metadata) for ob in obj["cells"]]

    def _parse_observation(self, obj):
        values = {
            k: np.array(v) if isinstance(v, list) else v
            for k, v in obj["values"].items()
        }
        if "prev_evaluation_date" in obj.keys():
            cell = IncrementalCell(
                period_start=datetime.datetime.strptime(
                    obj["period_start"], self.date_format
                ).date(),
                period_end=datetime.datetime.strptime(
                    obj["period_end"], self.date_format
                ).date(),
                evaluation_date=datetime.datetime.strptime(
                    obj["evaluation_date"], self.date_format
                ).date(),
                prev_evaluation_date=datetime.datetime.strptime(
                    obj["prev_evaluation_date"], self.date_format
                ).date(),
                values=values,
            )
        else:
            cell = CumulativeCell(
                period_start=datetime.datetime.strptime(
                    obj["period_start"], self.date_format
                ).date(),
                period_end=datetime.datetime.strptime(
                    obj["period_end"], self.date_format
                ).date(),
                evaluation_date=datetime.datetime.strptime(
                    obj["evaluation_date"], self.date_format
                ).date(),
                values=values,
            )
        return cell


class TriangleEncoder(json.JSONEncoder):
    def default(self, obj):
        return {"slices": [_slice_to_dict(slice_) for slice_ in obj.slices.values()]}


def _slice_to_dict(slice_: Triangle[Cell]) -> dict[str, Any]:
    return {
        **{
            k: v
            for k, v in slice_.cells[0].metadata.as_dict().items()
            if v is not None and v != {}
        },
        "cells": [_cell_to_dict(cell) for cell in slice_.cells],
    }


def _cell_to_dict(cell: Cell) -> dict[str, Any]:
    base_dict = {
        "period_start": cell.period_start.strftime("%Y-%m-%d"),
        "period_end": cell.period_end.strftime("%Y-%m-%d"),
        "evaluation_date": cell.evaluation_date.strftime("%Y-%m-%d"),
    }
    values_dict = {
        "values": {
            field: value.tolist() if isinstance(value, np.ndarray) else value
            for field, value in cell.values.items()
        },
    }
    if isinstance(cell, IncrementalCell):
        base_dict.update(
            {"prev_evaluation_date": cell.prev_evaluation_date.strftime("%Y-%m-%d")}
        )

    return {**base_dict, **values_dict}



__all__ = [
    "TriangleError",
    "TriangleEmptyError",
    "TriangleWarning",
    "DuplicateCellWarning",
]


class TriangleError(Exception):
    """Error specific to misuses of `Triangle`s."""


class TriangleEmptyError(TriangleError):
    """Error when a `Triangle` instance contains no cells or slices."""


class TriangleWarning(Warning):
    """Base class for all warnings from `Triangle`."""


class DuplicateCellWarning(TriangleWarning):
    """Warning when a Triangle contains multiple cells with the same coordinates."""



import datetime
from collections import defaultdict

from .. import date_utils
from ..base import Cell, CumulativeCell
from ..errors import TriangleError
from ..triangle import Triangle
from .basis import to_cumulative, to_incremental
from .summarize import summarize_cell_values

__all__ = [
    "aggregate",
]


def aggregate(
    triangle: Triangle[Cell],
    period_resolution: tuple[int, str] | None = None,
    eval_resolution: tuple[int, str] | None = None,
    period_origin: datetime.date = datetime.date(1999, 12, 31),
    eval_origin: datetime.date = datetime.date(1999, 12, 31),
    summarize_premium: bool = True,
) -> Triangle[Cell]:
    """Change the period and/or evaluation resolution of a triangle.

    Period resolution is defined as the length of each experience period in a triangle.
    Similarly, evaluation resolution is defined as the interval between each successive
    evaluation date in a triangle. Period resolution and evaluation are commonly equal to
    each other, but this is not a hard requirement.

    This aggregation implementation will not try to interpolate, apportion, or otherwise
    estimate triangle cells when data is insufficient to determine an exact answer. This
    means that if a triangle has two experience periods, `2020-01-01:2020-09-30` and
    `2020-10-01:2021-06-30`, aggregation to annual data is not possible.

    Aggregation over experience periods is equivalent to summing all values from all cells
    that are being merged into a new cell. No warnings are issued if not all cells have the
    same set of attributes, or if the new experience period isn't completely covered by
    merged cells. Depending on the semantics of your source data and intended application,
    this may not be advisable.

    Aggregation over evaluation dates is equivalent to filtering out evaluation dates.
    No effort is made to "carry forward" cells with gaps in their evaluation date history.
    The aggregation strategy is to sum fields over cells. This makes sense for fields such as
    paid loss, earned premium, and claim counts, but aggregation-by-summation doesn't make
    sense for fields such as loss ratio or average claim severity. This function does not
    attempt to infer an appropriate aggregation strategy based on the name of each field.

    Args:
        triangle: Triangle to aggregate.
        period_resolution: If supplied, the length of each experience perid. Format is a tuple
            with an integer number of units and a string specifying the unit. E.g.,
            `(7, "day")`, `(3, "months")`, `(1, "year")`. If missing, no experience period
            aggregation will be performed.
        eval_resolution: If supplied, the interval between each evaluation date. Format is
            the same as for `period_resolution`. If missing, no evaluation date aggregation
            will be performed.
        period_origin: A date that will be guaranteed to be a period start date (if there
            is any contemporaneous data). This argument is used for offseting experience
            periods temporally: e.g., for annual experience periods that start on July 1
            of each year, supply `period_origin=datetime.date(2000, 7, 1)` or similar.
        eval_origin: A date that will be guaranteed to be an evaluation date (if there
            is any contemporaneous data). See `period_origin` for details of semantics.
        summarize_premium: Whether or not to sum up premiums over aggregated cells. Defaults
            to True, set to False when summarizing over loss details over which premiums should
            not be summed.
    """
    cum_triangle = to_cumulative(triangle) if triangle.is_incremental else triangle
    agg_slices = []
    for slice_ in cum_triangle.slices.values():
        eval_agg_slice = _aggregate_eval(slice_, eval_resolution, eval_origin)
        agg_slice = _aggregate_period(
            eval_agg_slice, period_resolution, period_origin, summarize_premium
        )
        agg_slices.append(agg_slice)

    aggregate_triangle = (
        to_incremental(sum(agg_slices)) if triangle.is_incremental else sum(agg_slices)
    )

    return aggregate_triangle


def _aggregate_eval(triangle, eval_resolution, eval_origin):
    if eval_resolution is None:
        return triangle

    resolution = date_utils.standardize_resolution(eval_resolution)
    first_eval = triangle.evaluation_dates[0]
    last_eval = triangle.evaluation_dates[-1]

    # Move current_eval to on or before the first evaluation date in the slice
    current_eval = eval_origin
    while date_utils.resolution_delta(current_eval, resolution) < first_eval:
        current_eval = date_utils.resolution_delta(current_eval, resolution)
    while current_eval >= first_eval:
        current_eval = date_utils.resolution_delta(
            current_eval, resolution, negative=True
        )

    # Construct the set of all valid evaluations
    valid_evals = []
    current_eval = date_utils.resolution_delta(current_eval, resolution)
    while current_eval <= last_eval:
        valid_evals.append(current_eval)
        current_eval = date_utils.resolution_delta(current_eval, resolution)

    # Filter for only the valid evaluation dates
    return Triangle(
        [cell for cell in triangle.cells if cell.evaluation_date in valid_evals]
    )


# noinspection PyShadowingNames
def _aggregate_period(triangle, period_resolution, period_origin, summarize_premium):
    if period_resolution is None:
        return triangle

    resolution = date_utils.standardize_resolution(period_resolution)
    cells = sorted(triangle.cells, key=lambda cell: cell.coordinates)
    first_start = cells[0].period_start

    # Find the period_end immediately before the start of the aggregated triangle
    current_init = period_origin
    while date_utils.resolution_delta(current_init, resolution) < first_start:
        current_init = date_utils.resolution_delta(current_init, resolution)
    while current_init >= first_start:
        current_init = date_utils.resolution_delta(
            current_init, resolution, negative=True
        )

    # Map from old periods to new periods
    new_coordinate_cells = defaultdict(list)
    current_start = current_init + datetime.timedelta(days=1)
    current_end = date_utils.resolution_delta(current_init, resolution)
    for cell in cells:
        # We can step forward safely like this because we're sorted on period_start
        while current_end < cell.period_start:
            current_init = date_utils.resolution_delta(current_init, resolution)
            current_start = current_init + datetime.timedelta(days=1)
            current_end = date_utils.resolution_delta(current_init, resolution)

        # We know cell.period_start is in between current_start and current_end, we just
        # need to check for border crossings on the far side.
        if cell.period_end > current_end:
            raise TriangleError(
                f"Cannot aggregate periods to f{period_resolution} with "
                f"origin {period_origin}; triangle periods cross aggregate period bounds"
            )

        # Create a new cell to put in the coordinate pile
        new_cell = Cell(
            period_start=current_start,
            period_end=current_end,
            evaluation_date=cell.evaluation_date,
            values=cell.values,
            metadata=cell.metadata,
        )
        new_coordinate_cells[new_cell.coordinates].append(new_cell)

    # Aggregate the new cells
    new_cells = [
        CumulativeCell(
            period_start=cells[0].period_start,
            period_end=cells[0].period_end,
            evaluation_date=cells[0].evaluation_date,
            values=summarize_cell_values(cells, summarize_premium=summarize_premium),
            metadata=cells[0].metadata,
        )
        for cells in new_coordinate_cells.values()
    ]
    return Triangle(new_cells)



from dataclasses import replace
from typing import Any, Callable, Literal, Optional, get_args

import numpy as np
import toolz as tlz

from ..base import Cell, CellValue, CumulativeCell, IncrementalCell, Metadata
from ..errors import TriangleError
from ..triangle import Triangle

BLEND_METHOD_TYPE = Literal["mixture", "linear"]

__all__ = [
    "blend",
    "blend_samples",
    "blend_cells",
    "summarize",
    "summarize_cell_values",
    "split",
]


NON_LOSS_METRICS = {
    "earned_premium",
    "used_earned_premium",
    "earned_exposure",
    "written_premium",
    "written_exposure",
    "implied_atu",
    "bf_weight",
    "geometric_weight",
}

SUMMARIZE_DEFAULTS = {
    "paid_loss": lambda vd: _conforming_sum(vd["paid_loss"]),
    "reported_loss": lambda vd: _conforming_sum(vd["reported_loss"]),
    "incurred_loss": lambda vd: _conforming_sum(vd["incurred_loss"]),
    "reported_claims": lambda vd: _conforming_sum(vd["reported_claims"]),
    "open_claims": lambda vd: _conforming_sum(vd["open_claims"]),
    "closed_claims": lambda vd: _conforming_sum(vd["closed_claims"]),
    "closed_with_pay_claims": lambda vd: _conforming_sum(vd["closed_with_pay_claims"]),
    "reported_count": lambda vd: _conforming_sum(vd["reported_count"]),
    "open_count": lambda vd: _conforming_sum(vd["open_count"]),
    "closed_count": lambda vd: _conforming_sum(vd["closed_count"]),
    "closed_with_pay_count": lambda vd: _conforming_sum(vd["closed_with_pay_count"]),
    "earned_premium": lambda vd: _conforming_sum(vd["earned_premium"]),
    "used_earned_premium": lambda vd: _conforming_sum(vd["used_earned_premium"]),
    "earned_exposure": lambda vd: _conforming_sum(vd["earned_exposure"]),
    "written_premium": lambda vd: _conforming_sum(vd["written_premium"]),
    "written_exposure": lambda vd: _conforming_sum(vd["written_exposure"]),
    "implied_atu": lambda vd: _conforming_weighted_average(
        vd["implied_atu"], vd["reported_loss"]
    ),
    "bf_weight": lambda vd: _conforming_weighted_average(
        vd["bf_weight"], vd["reported_loss"]
    ),
    "geometric_weight": lambda vd: _conforming_weighted_average(
        vd["geometric_weight"], vd["reported_loss"]
    ),
    "log_industry_lr": lambda vd: _conforming_weighted_average(
        np.exp(vd["log_industry_lr"]), vd["earned_premium"], np.log
    ),
}

# For Bornhuetter-Ferguson method weighting
SUMMARIZE_DEFAULTS |= {
    f"{flavor}_loss_developed": lambda vd: _conforming_sum(
        vd[f"{flavor}_loss_developed"]
    )
    for flavor in ["incurred", "paid", "reported"]
}

SUMMARIZE_DEFAULTS |= {
    f"{flavor}_loss_prior": lambda vd: _conforming_sum(vd[f"{flavor}_loss_prior"])
    for flavor in ["incurred", "paid", "reported"]
}


# noinspection PyDefaultArgument
def summarize_cell_values(
    cells: list[Cell],
    agg_fns: Optional[dict[str, Callable]] = None,
    summarize_premium: bool = True,
) -> dict[str, CellValue]:
    """Combine all of the values in `cells` into a single dictionary.

    The semantics of combining each individual field in `values` is controlled by the
    `agg_fns` argument. Each summarization operation is passed a dictionary of all cell values
    so that fields whose aggregation depends on multiple values (i.e., loss ratio) can work
    correctly.

    If a key is present in some, but not all cells, default behavior is to pass in explicit
    `None` s for the missing keys into the aggregation function. Each aggregation function has the
    flexibility to decide whether to return a value in the presence of `None` s.
    """
    agg_fns = SUMMARIZE_DEFAULTS if agg_fns is None else SUMMARIZE_DEFAULTS | agg_fns

    value_keys = set([k for cell in cells for k in cell.values.keys()])
    for key in value_keys:
        if key.lower() not in agg_fns:
            raise TriangleError(f"Don't know how to aggregate `{key}` values")
    loss_keys = value_keys - NON_LOSS_METRICS
    non_loss_keys = value_keys & NON_LOSS_METRICS
    cell_raw_values = {
        key: [cell.values.get(key, None) for cell in cells] for key in value_keys
    }
    if summarize_premium:
        values_summaries = {
            key: agg_fns[key.lower()](cell_raw_values) for key in value_keys
        }
        return values_summaries

    else:
        non_loss_indices = _non_loss_distinct_indices(cells)
        cell_non_loss_values = {
            key: [elem for idx, elem in enumerate(values) if idx in non_loss_indices]
            for key, values in cell_raw_values.items()
        }

        loss_summaries = {
            key: agg_fns[key.lower()](cell_raw_values) for key in loss_keys
        }
        non_loss_summaries = {
            key: cell_non_loss_values[key][0] for key in non_loss_keys
        }

        return {
            **loss_summaries,
            **non_loss_summaries,
        }


def _non_loss_distinct_indices(cells: list[Cell]) -> list[int]:
    # per_occurrence_limit should work as a loss detail, we will keep the max limit
    all_limits = [
        cell.metadata.per_occurrence_limit
        for cell in cells
        if cell.metadata.per_occurrence_limit is not None
    ]
    max_limit = max(all_limits) if all_limits else None
    generic_metadata = [
        replace(cell.metadata, loss_details={}, per_occurrence_limit=max_limit)
        for cell in cells
    ]
    distinct_metadata = set()
    distinct_idxs = []
    for idx, meta in enumerate(generic_metadata):
        if meta not in distinct_metadata:
            distinct_metadata |= {meta}
            distinct_idxs.append(idx)
    return distinct_idxs


def _conforming_sum(values: list[CellValue]) -> CellValue:
    """Compute the conforming sum of a list of values."""
    total: CellValue = 0
    for val in values:
        if val is None:
            continue
        # If we're summing two arrays, check shape
        if not np.isscalar(total) and not np.isscalar(val) and total.shape != val.shape:
            raise ValueError(
                f"Cannot sum arrays with shapes {total.shape} and {val.shape}"
            )
        total += val
    return total


def _conforming_weighted_average(
    values: list[CellValue],
    weights: list[CellValue],
    post_transform: Callable = lambda cell: cell,
) -> CellValue:
    """Compute the conforming weighted average of a list of values."""
    total: CellValue = 0
    for val, weight in zip(values, weights):
        if val is None:
            continue
        # If we're averaging two arrays, check shape
        if not np.isscalar(total) and not np.isscalar(val) and total.shape != val.shape:
            raise ValueError(
                f"Cannot average arrays with shapes {total.shape} and {val.shape}"
            )
        total += val * weight
    return post_transform(
        total / sum(weight for weight in weights if weight is not None)
    )


def blend_samples(
    values: list[CellValue],
    weights: list[float] = None,
    method: BLEND_METHOD_TYPE = "mixture",
    seed: int = None,
) -> CellValue:
    if weights is None:
        weights = np.repeat(1 / len(values), len(values))
    else:
        weights = np.asarray(weights)

    if len(weights) != len(values):
        raise ValueError(
            "The number of weights does not match the number of cell values"
        )

    if method == "mixture":
        return _mixture_blend(values, weights, seed)
    elif method == "linear":
        return _linear_blend(values, weights)
    else:
        raise ValueError(f"Unexpected blending method `{method}`.")


def _linear_blend(values: list[CellValue], weights: list = None):
    """Combine cell values by linear blending."""

    values = [[v] if isinstance(v, (float, int)) else v for v in values]
    S = max(len(v) for v in values)
    M = len(values)

    if np.ndim(weights) == 1:
        weights = np.array(weights).reshape((M, 1))

    # Create a 2-dimensional array of values ready for matrix multiplication
    matched_values = np.empty((S, M))
    for idx, val in enumerate(values):
        if len(val) != S and len(val) != 1:
            raise ValueError(
                f"Linear blending requires values of length 1 or {S}, not {len(val)}."
            )
        if len(val) == 1:
            matched_values[:, idx] = np.tile(val, S)
        else:
            matched_values[:, idx] = val

    blend = matched_values @ weights
    return blend.reshape((S,))


def _mixture_blend(
    values: list[CellValue], weights: list = None, seed: int = None
) -> CellValue:
    """Combine a set of samples of size M, each containing an S x 1 vector of samples,
    by a set of weights."""

    if round(sum(weights), 6) != 1:
        raise ValueError(f"Sum of weight parameters in {weights} must be exactly 1")

    np.random.seed(seed)

    M = len(values)
    S = np.shape(values[0])[0]

    blend_idx = np.random.choice(range(M), S, p=weights)
    new_cell_values = np.empty((S,))
    for idx, v in enumerate(values):
        new_cell_values[blend_idx == idx] = v[blend_idx == idx]

    return new_cell_values


def blend_cells(
    cells: list[Cell],
    weights: list[float],
    method: BLEND_METHOD_TYPE,
    seed: Optional[int],
) -> Cell:
    location_str = (
        f"period {cells[0].period}, evaluation date {cells[0].evaluation_date}"
    )

    # Ensure that all cells have the same set of fields
    fields = set(cells[0].values.keys())
    for cell in cells[1:]:
        if set(cell.values.keys()) != fields:
            raise ValueError("Cannot blend cells with an inconsistent set of fields")

    # Blend each field in values, one by one
    clean_values = {}
    for field in fields:
        field_vals = [cell[field] for cell in cells]

        # check that all values have the same class for mixture blending
        if method == "mixture" and not all(
            [isinstance(val, type(field_vals[0])) for val in field_vals[1:]]
        ):
            raise TypeError(
                f"Attempted to mixture blend triangle with inconsistent field types at {location_str}"
            )

        # Check that scalar values are all the same for mixture blending
        # Linear blending can blend distinct scalars
        if method == "mixture" and np.isscalar(field_vals[0]):
            if any([val != field_vals[0] for val in field_vals[1:]]):
                raise ValueError(
                    f"Attempted to mixture blend triangle with inconsistent scalar values at {location_str}"
                )
            clean_values[field] = field_vals[0]
        # otherwise, do method blending
        else:
            clean_values[field] = blend_samples(field_vals, weights, method, seed)

    return cells[0].replace(values=clean_values)


def blend(
    triangles: list[Triangle],
    weights: list[float] | dict[str, np.ndarray] | None = None,
    method: BLEND_METHOD_TYPE = "mixture",
    seed: int | None = None,
) -> Triangle:
    """Return a weighted blend of triangles by their values.

    There are two distinct methods of blending.
    Mixture blending samples cell values in
    proportion to their weight. Linear blending
    is a linear weighted average of the cell values.
    Mixture blending only applies to distributions of
    cell values with the same dimensions, but linear
    blending can be applied between distributions
    and scalars alike.

    Args:
        triangles: The list of triangles to blend.
        weights: The optional list or dictionary of weights.
            Cell-wise weights are passed as a dictionary.
        method: The blending method to use. Can be one of
            ['mixture', 'linear']. Defaults to `mixture`.
        seed: The seed to use for mixture blending.

    Returns:
        A blended triangle.
    """

    if not isinstance(triangles, list):
        raise TypeError(
            f"Can only blend a list of triangles, not a type {type(triangles)}."
        )

    if len(triangles) <= 1 and weights[0] != 1.0:
        raise ValueError(
            f"Blending single triangles requires weight = [1.0] not {weights}."
        )

    if method.lower() not in get_args(BLEND_METHOD_TYPE):
        raise ValueError(
            f"Blending method `{method}` not found. Available methods are {get_args(BLEND_METHOD_TYPE)}.`"
        )
    else:
        method = method.lower()

    # Make sure that all of the triangles are the same length and have the same cell type
    # We'll check to make sure that all of the cell indices and metadata match later
    n_cells = len(triangles[0])
    if any([len(tri) != n_cells for tri in triangles[1:]]):
        raise ValueError("All triangles in blend must have the same number of elements")
    if any(
        [type(tri.cells[0]) != type(triangles[0].cells[0]) for tri in triangles[1:]]  # noqa: E721
    ):
        raise ValueError("All triangles in blend must have the same cell type")

    if triangles[0].is_incremental:
        index_triangles = [
            {
                (
                    cell.metadata,
                    cell.period,
                    cell.evaluation_date,
                    cell.prev_evaluation_date,
                ): cell
                for cell in tri
            }
            for tri in triangles
        ]
    else:
        index_triangles = [
            {(cell.metadata, cell.period, cell.evaluation_date): cell for cell in tri}
            for tri in triangles
        ]

    # re-structure weights dependent on input type
    if weights is None:
        weight_list = [None] * n_cells
    elif isinstance(weights, dict):
        weights_2d = {k: np.atleast_2d(v) for k, v in weights.items()}
        weight_list = [weight for weight in np.concatenate(list(weights_2d.values())).T]
        if len(weight_list) != 1 and len(weight_list) != n_cells:
            raise ValueError(
                "Dimensions of `weights` do not match those of `triangles`. Either a single "
                "set of weights should be supplied that will be applied to all cells "
                "in `triangles`, or exactly one set of weights for each cell in `triangles` "
                "should be supplied for cell-wise blending."
            )
        if len(weight_list) == 1:
            weight_list = weight_list * n_cells
    elif isinstance(weights, list):
        weight_list = [weights] * n_cells
    else:
        raise TypeError(
            f"`weights` passed to bermuda.utils.blend has unknown type {type(weights)}."
        )

    blended_cells = []
    for ndx, cell_weights in zip(index_triangles[0], weight_list):
        try:
            ndx_cells = [ndx_tri[ndx] for ndx_tri in index_triangles]
        except KeyError:
            raise ValueError(
                "All triangles in a blend must have an identical set of coordinates"
            )
        blended_cells.append(blend_cells(ndx_cells, cell_weights, method, seed))

    return Triangle(blended_cells)


def split(triangle: Triangle, detail_keys: list[str]) -> dict[Any, Triangle]:
    """Turn a Triangle into a dictionary of Triangles grouped by specified detail keys."""

    def key_fn(cell):
        detail_vals = [cell.metadata.details.get(key, None) for key in detail_keys]
        return tuple(detail_vals)

    grouped_cells = tlz.groupby(key_fn, triangle.cells)
    return tlz.valmap(Triangle, grouped_cells)


def summarize(
    triangle: Triangle,
    summary_fns: Optional[dict[str, Callable]] = None,
    summarize_premium: bool = True,
) -> Triangle:
    """Aggregate a Triangle across metadata.

    For example, given a triangle with state and coverage-level detail, this would generate a
    triangle where all cells with the same coordinates are merged, and any metadata that varies
    over the triangle would be dropped.

    Args:
        triangle: The single triangle to summarize.
        summary_fns: A dictionary of functions to use to summarize each field in the triangle. If
            `None`, a default set of functions is supplied. If specified, the dictionary is added to the default set of summary functions, which makes it possible to easily add functions for custom fields without affecting defaults.
        summarize_premium: Whether to summarize the premium field. If False, the premium field
            will not be summed accross cells, but will be copied from the first matching cell
            in the triangle.

    Returns:
        A summarized triangle with a single cell for each unique combination of coordinates.
    """
    # noinspection PyShadowingNames
    common_metadata = _metadata_gcd(triangle)

    summary_cells = []
    if triangle.is_incremental:
        grouped_cells = tlz.groupby(
            lambda cell: (cell.period, cell.evaluation_date, cell.prev_evaluation_date),
            triangle.cells,
        )
        for (period, eval_date, prev_eval_date), cells in grouped_cells.items():
            summary_cells.append(
                IncrementalCell(
                    period_start=period[0],
                    period_end=period[1],
                    evaluation_date=eval_date,
                    prev_evaluation_date=prev_eval_date,
                    metadata=common_metadata,
                    values=summarize_cell_values(cells, summary_fns),
                )
            )
    else:
        grouped_cells = tlz.groupby(
            lambda cell: (cell.period, cell.evaluation_date), triangle.cells
        )
        for (period, eval_date), cells in grouped_cells.items():
            summary_cells.append(
                CumulativeCell(
                    period_start=period[0],
                    period_end=period[1],
                    evaluation_date=eval_date,
                    metadata=common_metadata,
                    values=summarize_cell_values(cells, summary_fns, summarize_premium),
                )
            )

    return Triangle(summary_cells)


def _metadata_gcd(triangle: Triangle[Cell]) -> Metadata:
    if not triangle.has_consistent_risk_basis:
        raise TriangleError("All cells in `summarize` must have the same risk basis")
    if not triangle.has_consistent_currency:
        raise TriangleError("All cells in `summarize` must have the same currency")

    return Metadata(
        risk_basis=triangle.cells[0].risk_basis,
        currency=triangle.cells[0].currency,
        country=_metadata_attr_gcd(triangle, "country"),
        per_occurrence_limit=_metadata_attr_gcd(triangle, "per_occurrence_limit"),
        loss_definition=_metadata_attr_gcd(triangle, "loss_definition"),
        reinsurance_basis=_metadata_attr_gcd(triangle, "reinsurance_basis"),
        details=_details_gcd([cell.metadata.details for cell in triangle.cells]),
        loss_details=_details_gcd(
            [cell.metadata.loss_details for cell in triangle.cells]
        ),
    )


def _details_gcd(dicts: list[dict[str, Any]]) -> dict[str, Any]:
    # Find the keys that are present in all of the dicts
    common_keys = set(dicts[0].keys())
    for dct in dicts[1:]:
        common_keys &= set(dct.keys())

    # Check if those keys have the same values in all of the dicts
    common_values = {}
    for key in common_keys:
        first_value = dicts[0][key]
        consistent_value = True
        for dct in dicts[1:]:
            if dct[key] != first_value:
                consistent_value = False
                break
        if consistent_value and first_value is not None:
            common_values[key] = first_value

    return common_values


def _metadata_attr_gcd(triangle: Triangle[Cell], attr_name: str):
    first_val = getattr(triangle.cells[0].metadata, attr_name)
    if all([getattr(cell.metadata, attr_name) == first_val for cell in triangle]):
        return first_val
    return None



import struct

__all__ = [
    "MAGIC",
    "VERSION",
    "STRING",
    "INT",
    "FLOAT",
    "BOOL",
    "NONE",
    "DATE",
    "INT_ARRAY",
    "FLOAT_ARRAY",
    "DICT_END",
    "METADATA",
    "CELL",
    "CUMULATIVE_CELL",
    "INCREMENTAL_CELL",
]

# Useful constants
MAGIC = struct.pack("<L", 0x0136AF)
VERSION = struct.pack("B", 0x01)

# Flags for basic data types
STRING = struct.pack("B", 0x80)
INT = struct.pack("B", 0x81)
FLOAT = struct.pack("B", 0x82)
BOOL = struct.pack("B", 0x83)
NONE = struct.pack("B", 0x84)
DATE = struct.pack("B", 0x85)
INT_ARRAY = struct.pack("B", 0x86)
FLOAT_ARRAY = struct.pack("B", 0x87)
DICT_END = struct.pack("B", 0x88)

# Flags for Higher-order data types
METADATA = struct.pack("B", 0x10)
CELL = struct.pack("B", 0x11)
CUMULATIVE_CELL = struct.pack("B", 0x12)
INCREMENTAL_CELL = struct.pack("B", 0x13)



from dataclasses import replace

from ..base import Cell
from ..triangle import Triangle

__all__ = [
    "convert_to_dollars",
    "convert_currency",
]


# Define the list of fields that are currency-denominated and that therefore need to be
# converted when changing currencies
CURRENCY_FIELDS = [
    "earned_premium",
    "used_earned_premium",
    "written_premium",
    "paid_loss",
    "reported_loss",
    "incurred_loss",
]


# These are rough defaults and should almost always be overridden for a real analysis
DEFAULT_EXCHANGE_RATES = {
    "EUR": 1.10,
    "GBP": 1.31,
}


def convert_to_dollars(
    triangle: Triangle, exchange_rates: dict[str, float] | None = None
) -> Triangle:
    """Convert all currency-denominated fields in a Triangle to USD.

    See `convert_currency` for more details on the conversion process.

    Arguments:
        triangle: A triangle to convert to dollars.
        exchange_rates: A dictionary that maps currencies to currency-to-dollar exchange rates.
    Returns:
        A triangle with all cells denominated in dollars.
    """
    if exchange_rates is None:
        exchange_rates = DEFAULT_EXCHANGE_RATES
    return convert_currency(triangle, "USD", exchange_rates)


def convert_currency(
    triangle: Triangle, target_currency: str, exchange_rates: dict[str, float]
) -> Triangle:
    """Convert all currency-denominated fields in a Triangle to a target currency.

    Non-currency fields (e.g., claim counts, policy counts) will not be altered.

    Arguments:
        triangle: A triangle to convert to a single currency.
        target_currency: The currency that all values will be expressed in, e.g. `USD` or `EUR`.
        exchange_rates: A dictionary with exchange rates from other currencies to the target
            currency. For example, to convert a triangle with GBP-denominated losses to USD
            with an assumed exchange rate of 1.0 GBP = 1.4 USD, then `exchange_rates` would
            be `{"GBP": 1.4}`.
    Returns:
        A triangle with all cells denominated in the target currency.
    """
    converted_cells = []
    for metadata, tri_slice in triangle.slices.items():
        if metadata.currency is None:
            raise ValueError(
                "Every slice in the input triangle must have a defined currency"
            )
        elif metadata.currency == target_currency:
            converted_cells += tri_slice.cells
        elif metadata.currency not in exchange_rates:
            raise ValueError(
                f"Must supply an exchange rate for currency `{metadata.currency}`"
            )
        else:
            exchange_rate = exchange_rates[metadata.currency]
            for cell in tri_slice:
                converted_cells.append(
                    _convert_cell_currency(cell, exchange_rate, target_currency)
                )

    return Triangle(converted_cells)


def _convert_cell_currency(
    cell: Cell, exchange_rate: float, target_currency: str
) -> Cell:
    converted_values = {
        k: v * exchange_rate if k in CURRENCY_FIELDS else v
        for k, v in cell.values.items()
    }

    return cell.replace(
        values=converted_values,
        metadata=replace(cell.metadata, currency=target_currency),
    )



from __future__ import annotations

import datetime

from ..base import CumulativeCell, Metadata
from ..date_utils import add_months
from ..triangle import Triangle
from .data_frame_output import triangle_to_wide_data_frame


def triangle_to_chain_ladder(
    triangle: Triangle,
) -> chainladder.core.triangle.Triangle:  # noqa: F821
    """Convert a Triangle to a chainladder Triangle object.

    Args:
        triangle (Triangle): The Bermuda Triangle to convert.

    Returns:
        chainladder.core.triangle.Triangle: The chainladder Triangle object.
    """
    try:
        import chainladder as cl
    except ImportError:
        raise ImportError(
            "The chainladder package is required to convert a Triangle to a ChainLadder object"
        )
    df = triangle_to_wide_data_frame(triangle)
    index = set(df.columns) - set(
        triangle.fields + ["period_start", "period_end", "evaluation_date"]
    )
    cl_triangle = cl.Triangle(
        df,
        origin="period_start",
        development="evaluation_date",
        columns=triangle.fields,
        origin_format="%Y-%m-%d",
        development_format="%Y-%m-%d",
        cumulative=not triangle.is_incremental,
        index=list(index),
    )
    return cl_triangle


def chain_ladder_to_triangle(
    chain_ladder: chainladder.core.triangle.Triangle,  # noqa: F821
    base_metadata: Metadata | None = None,
) -> Triangle:
    """Create a Triangle from a chainladder Triangle object.

    Anything in the chainladder index will become details in the Bermuda Triangle.

    Args:
        chain_ladder (chainladder.core.triangle.Triangle): The chainladder Triangle object.
        base_metadata (Metadata): The base metadata to apply to all cells. Defaults to None.

    Returns:
        Triangle: The Bermuda Triangle.
    """
    if base_metadata is None:
        base_metadata = Metadata()
    if not chain_ladder.is_cumulative:
        raise ValueError("Only cumulative ChainLadder triangles are supported")
    cl_df = chain_ladder.val_to_dev().to_frame()
    origin_resolution = {"Y": 12, "Q": 3, "M": 1}[chain_ladder.origin_grain]
    if chain_ladder.shape[0] == 1:
        df_long = cl_df.reset_index().melt(
            id_vars="index", var_name="dev_lag", value_name="values"
        )
        cells = []
        for _, row in df_long.dropna(subset=["values"]).iterrows():
            period_end = add_months(
                row["index"], origin_resolution
            ) - datetime.timedelta(days=1)

            cell_values = {chain_ladder.columns[0]: row["values"]}

            cell = CumulativeCell(
                period_start=row["index"].date(),
                period_end=period_end,
                evaluation_date=add_months(
                    period_end, row["dev_lag"] - origin_resolution
                ),
                values=cell_values,
                metadata=base_metadata,
            )
            cells.append(cell)
        return Triangle(cells)
    # group by index and iterate over groups
    index_names = chain_ladder.key_labels
    cells = []
    for group_id, group_df in cl_df.groupby(level=index_names):
        details = {k: v for k, v in zip(index_names, group_id)}
        for _, row in group_df.iterrows():
            period_end = add_months(
                row["origin"], origin_resolution
            ) - datetime.timedelta(days=1)

            cell_values = row.drop(["origin", "development"]).to_dict()

            cell = CumulativeCell(
                period_start=row["origin"].date(),
                period_end=period_end,
                evaluation_date=add_months(
                    period_end, row["development"] - origin_resolution
                ),
                values=cell_values,
                metadata=Metadata(**{**base_metadata.as_dict(), "details": details}),
            )
            cells.append(cell)

    return Triangle(list(cells))



from __future__ import annotations

import dataclasses
import datetime
from collections import namedtuple
from typing import Any, Callable, Tuple, Union, get_args

import numpy as np

from ..date_utils import calculate_dev_lag, month_to_id
from .metadata import METADATA_COMMON_ATTRIBUTES, Metadata

__all__ = [
    "CellValue",
    "Cell",
    "values_eq",
    "format_value",
]


_constr_coordinates = namedtuple(
    "coordinates", ["period_start", "period_end", "evaluation_date"]
)
CellValue = Union[float, int, np.ndarray, np.float64, np.int64, None]


class Cell(object):
    """Base class for all elements contained within `Triangle`s.

    Attributes:
        period_start: The beginning of the experience period.
        period_end: The end of the experience period (as a closed interval).
        evaluation_date: The date when the information was known.
        values: dictionary with key-value pairs of fields and values.
            For example: `{"paid_loss": 123400, "earned_premium": 2345000}`
        metadata: Other metadata about the `Cell`.
    """

    def __init__(
        self,
        period_start: datetime.date,
        period_end: datetime.date,
        evaluation_date: datetime.date,
        values: dict[str, CellValue],
        metadata: Metadata = None,
        _skip_validation: bool = False,
    ):
        if not _skip_validation:
            # Simple input type checks
            if not isinstance(period_start, datetime.date):
                raise TypeError("`period_start` must be a `datetime.date`")
            if not isinstance(period_end, datetime.date):
                raise TypeError("`period_end` must be a `datetime.date`")
            if not isinstance(evaluation_date, datetime.date):
                raise TypeError("`evaluation_date` must be a `datetime.date`")
            if not isinstance(metadata, (Metadata, type(None))):
                raise TypeError("`metadata` must be a `Metadata` object (or `None`)")

            # Check that all values are of the correct type (or can be coerced to the correct type)
            if not isinstance(values, dict):
                raise TypeError("`values` must be a `dict`")
            for name, val in values.items():
                if not isinstance(val, get_args(CellValue)):
                    raise TypeError(
                        f"Field {name} has unexpected value type {val.__class__.__name__}"
                    )
                if isinstance(val, np.ndarray):
                    # See binary_output.py line 212
                    if not isinstance(val.dtype, (np.int64, np.float64)):
                        try:
                            val = val.astype(np.float64)
                        except ValueError as e:
                            raise ValueError(
                                f"""
                            Error: The array in {name} is not of dtype `np.int64` or `np.float64`,
                            nor is it coercible to `np.float64`. Outputting to binary format is only
                            compatible with 64-bit numeric types.
                            """
                            ) from e

            # Check for date consistency
            if period_end < period_start:
                raise ValueError("`period_end` must be >= `period_start`")
            if evaluation_date < period_start:
                raise ValueError(
                    f"`evaluation_date` ({evaluation_date}) must be >= `period_start` ({period_start})."
                )
            if evaluation_date == datetime.date.max:
                raise ValueError(
                    f"`evaluation_date` ({evaluation_date}) must be < `datetime.date.max`."
                )

        # Store dates as datetime.date objects (dropping time component, if present)
        self._period_start = datetime.date(
            period_start.year, period_start.month, period_start.day
        )
        self._period_end = datetime.date(
            period_end.year, period_end.month, period_end.day
        )
        self._evaluation_date = datetime.date(
            evaluation_date.year, evaluation_date.month, evaluation_date.day
        )

        self._values = values if values is not None else {}
        self._metadata = metadata if metadata is not None else Metadata()

    # === Accessors =============================================================================
    # Shorthand implementation for properties on all basic fields
    period_start = property(lambda self: self._period_start)
    period_end = property(lambda self: self._period_end)
    evaluation_date = property(lambda self: self._evaluation_date)
    values = property(lambda self: self._values)
    metadata = property(lambda self: self._metadata)
    risk_basis = property(lambda self: self._metadata.risk_basis)
    country = property(lambda self: self._metadata.country)
    currency = property(lambda self: self._metadata.currency)
    reinsurance_basis = property(lambda self: self._metadata.reinsurance_basis)
    loss_definition = property(lambda self: self._metadata.loss_definition)
    per_occurrence_limit = property(lambda self: self._metadata.per_occurrence_limit)
    details = property(lambda self: self._metadata.details)
    loss_details = property(lambda self: self._metadata.loss_details)

    @property
    def period(self) -> Tuple[datetime.date, datetime.date]:
        return self._period_start, self._period_end

    @property
    def period_length(self) -> int:
        """The length of the cell experience period, in months."""
        return month_to_id(self._period_end) - month_to_id(self._period_start) + 1

    @property
    def coordinates(self) -> Tuple[datetime.date, datetime.date, datetime.date]:
        """Location of the cell: period_start, period_end, evaluation_date."""
        return _constr_coordinates(
            self._period_start, self._period_end, self._evaluation_date
        )

    def dev_lag(self, unit: str = "months") -> Union[float, int, datetime.timedelta]:
        """Compute the development lag of the cell in the specified units.

        Args:
            unit (str): One of 'month', 'day', 'timedelta'.  'month' returns dev_lag as
                a float number of months, day returns dev_lag as an integer number
                of days, and 'timedelta' returns a datetime.timedelta object.
        """
        return calculate_dev_lag(self._period_end, self._evaluation_date, unit)

    # === Dunder methods ========================================================================

    def __hash__(self):
        # Numpy arrays aren't hashable, so manually handle the values dictionary
        value_hashes = []
        for k, v in self._values.items():
            if isinstance(v, np.ndarray):
                item_hash = hash((k, tuple(v)))
            else:
                item_hash = hash((k, v))
            value_hashes.append(item_hash)

        return hash(
            (
                # Including class name to ensure cumulative and incremental are distinct
                self.__class__.__name__,
                self._period_start,
                self._period_end,
                self._evaluation_date,
                self._metadata,
                # Sorting value hashes here to ensure insensitivity to
                # key insertion order
                tuple(sorted(value_hashes)),
            )
        )

    def __getitem__(self, key: str) -> Any:
        try:
            return self._values[key]
        except KeyError:
            raise KeyError(
                f"Field `{key}` does not exist in this `{self.__class__.__name__}`"
            )

    def __contains__(self, key: str) -> bool:
        return key in self._values

    def __eq__(self, other: Cell) -> bool:
        return (
            # Defining class equality this way to handle cases where Cells are being
            # compared to CumulativeCells or vice-versa.
            (isinstance(self, other.__class__) or isinstance(other, self.__class__))
            and self._period_start == other._period_start
            and self._period_end == other._period_end
            and self._evaluation_date == other._evaluation_date
            and self._metadata == other._metadata
            and values_eq(self._values, other._values)
        )

    def __lt__(self, other: Cell) -> bool:
        """Required so that lists of cells can be sortable.
        Comparison is based on (period_start, period_end, evaluation_date) tuple.
        """
        return (
            self._metadata,
            self._period_start,
            self._period_end,
            self._evaluation_date,
        ) < (
            other._metadata,
            other._period_start,
            other._period_end,
            other._evaluation_date,
        )

    def __repr__(self) -> str:
        attr_strings = []
        for name, val in self.__dict__.items():
            attr_name = name[1:] if name[0] == "_" else name
            attr_strings.append(f"{attr_name}={repr(val)}")
        return self.__class__.__name__ + "(" + ", ".join(attr_strings) + ")"

    def _repr_html_(self) -> str:
        metadata_rows = []
        for name, value in self.metadata.as_flat_dict().items():
            if value is not None:
                metadata_rows.append(f"<tr><th> {name} </th> <td> {value} </td></tr>")

        value_rows = []
        for name, value in self.values.items():
            if isinstance(value, np.ndarray):
                # noinspection PyTypeChecker
                value_summary = f"~{format_value(np.mean(value))} (Size {value.size})"
            else:
                value_summary = format_value(value)
            value_rows.append(f"<tr><th> {name} </th> <td> {value_summary} </td></tr>")

        table = [
            "<table>",
            "  <thead>",
            f"    <tr><th> <h4><b>{self.__class__.__name__}</b></h4> </th></tr>",
            "  </thead>",
            "  <tbody>",
            f"    <tr><th> Period Start: </th><td> {self.period_start} </td></tr>",
            f"    <tr><th> Period End: </th><td> {self.period_end} </td></tr>",
            f"    <tr><th> Evaluation Date: </th><td> {self.evaluation_date} </td></tr>",
            f"    <tr><th> Development Lag: </th><td> {self.dev_lag()} months </td></tr>",
            '    <tr><th colspan=2 style="text-align:center"> Values </th></tr>',
            *value_rows,
            '    <tr><th colspan=2 style="text-align:center"> Metadata </th></tr>',
            *metadata_rows,
            "  </tbody></table>",
        ]
        return "\n".join(table)

    # === Transformer methods ===================================================================

    def _base_replace(self, **definitions) -> Cell:
        """Override values of arbitrary fields in a non-destructive fashion."""
        attrs = {k[1:]: v for k, v in self.__dict__.items()}
        attrs.update(definitions)
        return self.__class__(**attrs)

    def replace(self, **definitions: Union[Callable[[Cell], Any], Any]) -> Cell:
        """Override values of arbitary fields with constants or functions.
        Calls _base_replace internally. Only validates the final cell once
        all replacements have been made.
        """
        cell = self
        for name, func_or_val in definitions.items():
            value = func_or_val(cell) if callable(func_or_val) else func_or_val
            definitions = {name: value}
            cell = cell._base_replace(**definitions, _skip_validation=True)
        return cell._base_replace(_skip_validation=False)

    def select(self, keys: list[str]) -> Cell:
        """Retain only the designated keys in `values`."""
        new_values = {k: v for k, v in self.values.items() if k in keys}
        return self.replace(values=new_values)

    def derive_fields(self, **definitions: Union[Callable[[Cell], Any], Any]) -> Cell:
        """Add new derived fields to a cell.

        Arguments:
            **definitions: Values are functions that accept a `Cell` and return a value,
                and keys are the names that the function values will be assigned to.
        Examples:
            >>> ob = Cell(
            ...     period_start=0,
            ...     period_end=1,
            ...     values={
            ...         "loss": 10,
            ...         "premium": 50,
            ...     },
            ... )
            >>> new_ob = ob.derive_fields(
            ...     loss_ratio=lambda c: c[
            ...         "loss"
            ...     ]
            ...     / c["premium"]
            ... )
            >>> new_ob.values
            {"loss": 10, "premium": 50, "loss_ratio": 0.2}
        """
        cell = self
        for name, func_or_val in definitions.items():
            value = func_or_val(cell) if callable(func_or_val) else func_or_val
            definitions = {name: value}
            cell = cell.replace(values={**cell.values, **definitions})
        return cell

    def derive_metadata(self, **definitions: Union[Callable[[Cell], Any], Any]) -> Cell:
        """Add new derived attributes to a Cell's metadata.

        Arguments:
            **definitions: Values are plain values or functions that accept a `Cell` and
                return a value. If keys are top-level attributes in metadata, their
                values will replace the existing values of the top-level attributes.
                Otherwise, the key-value pair will be added to the `details` field of
                the metadata.
        """
        cell = self
        for name, func_or_val in definitions.items():
            value = func_or_val(cell) if callable(func_or_val) else func_or_val
            if name in METADATA_COMMON_ATTRIBUTES:
                new_metadata = dataclasses.replace(cell.metadata, **{name: value})
            else:
                new_details = {**cell.metadata.details, name: value}
                new_metadata = dataclasses.replace(cell.metadata, details=new_details)
            cell = cell._base_replace(metadata=new_metadata)
        return cell

    def to_record(self, dev_lag_unit="month") -> dict[str, Any]:
        """Return dict representation of Cell."""
        return {
            "period_start": self._period_start,
            "period_end": self._period_end,
            "evaluation_date": self._evaluation_date,
            "dev_lag": self.dev_lag(dev_lag_unit),
            **self._metadata.as_flat_dict(),
            **self._values,
        }

    def add_statics(self, source_cell, fields) -> Cell:
        """Add new static fields (e.g., `earned_premium`) to a cell."""
        static_fields = {k: v for k, v in source_cell.values.items() if k in fields}
        return self.replace(values={**self._values, **static_fields})


def values_eq(val1, val2):
    # Utility function to see if two objects of type dict[str, np.ndarray] are identical.
    if not sorted(val1.keys()) == sorted(val2.keys()):
        return False
    for k in val1.keys():
        if not np.array_equal(val1[k], val2[k]):
            return False
    return True


def format_value(val: Union[float, int], sig_figs: int = 4) -> str:
    if sig_figs < 3:
        raise Exception("sig_figs must be at least 3")
    # Prep some handy properties of val
    sign = "-" if val < 0 else ""
    aval = abs(val)
    scale = -np.inf if val == 0.0 else np.floor(np.log10(aval))
    exponent = 3 * (scale // 3)
    # If it's an integer under sig_figs, just return the integer
    if (isinstance(val, int) or val == 0.0) and scale <= sig_figs:
        return str(val)
    # Special handling of values between -1.0 and 1.0
    if scale < 0:
        zeros = int(abs(-scale)) - 1
        digits = int(round(aval * 10 ** (zeros + sig_figs)))
        return sign + "0." + ("0" * zeros) + str(digits)
    # Figure out the suffix
    if exponent == 0:
        suffix = ""
    elif exponent == 3:
        suffix = "K"
    elif exponent == 6:
        suffix = "M"
    elif exponent == 9:
        suffix = "B"
    elif exponent == 12:
        suffix = "T"
    else:
        # Switch to engineering notation for extreme values
        suffix = "e" + str(int(exponent))
    mantissa = aval / (10**exponent)
    round_places = int(sig_figs - (scale - exponent + 1))
    return sign + str(round(mantissa, round_places)) + suffix



import numpy as np
import toolz as tlz

from ..errors import TriangleError
from ..triangle import Triangle

__all__ = [
    "add_statics",
    "array_from_field",
    "array_sizes",
    "array_size",
]


def add_statics(
    triangle: Triangle,
    source: Triangle,
    statics: list[str] = ["earned_premium", "earned_exposure"],
) -> Triangle:
    """Add "static" fields to a `Triangle`.
    Some fields (for example, accident-period earned premium) can be considered as fixed
    with respect to development lag. This method takes static fields from one triangle and
    adds them to another triangle. For example, if we have a triangle of predicted `paid_loss`
    and another disjoint triangle with observed `paid_loss`, we can fill in `earned_premium`
    values in the former triangle with values from the latter triangle.
    Args:
        source: A `Triangle` that will be used to fill in static fields.
        statics: A list of field names to fill in.
    Returns:
        A `Triangle` with static fields filled in from `source`.
    """

    rich_cells = []
    source_slices = source.slices
    for key, slc in triangle.slices.items():
        source_slice = source_slices.get(key, None)
        if source_slice is not None:
            rich_cells += _add_statics_slice(slc, source_slice, statics)
        else:
            rich_cells += slc.cells
    return Triangle(rich_cells)


def _add_statics_slice(slice, source_slice, statics):
    source_indexed = tlz.valmap(
        lambda row: sorted(row, key=lambda ob: ob.evaluation_date)[-1],
        tlz.groupby(lambda cell: cell.period, source_slice.cells),
    )

    rich_cells = []
    for cell in slice:
        source_cell = source_indexed.get(cell.period, None)
        if source_cell is not None:
            rich_cells.append(cell.add_statics(source_cell, statics))
        else:
            rich_cells += [cell]

    return rich_cells


def array_from_field(triangle: Triangle, field: str) -> np.ndarray:
    """Return an array with cell-level values from the triangle.

    Args:
        tri: Triangle to pull the cell values from.
        field: Name of the field to turn into an array.
    Returns:
        An array of dimension (N, K) where N is the number of cells in the source triangle
            and K is the array_size of the field.
    """

    if not all([field in cell for cell in triangle]):
        raise TriangleError(
            f"Cannot construct an array of {field} because the field is not present in every cell"
        )

    target_size = array_size(triangle)
    consistent_values = []
    for cell in triangle.cells:
        value = np.array(cell[field])
        if value.size != target_size:
            value = np.full((target_size,), value)
        consistent_values.append(value)
    return np.array(consistent_values)


def array_sizes(triangle: Triangle) -> list[int]:
    """Get the ordered set of all distinct non-scalar array sizes in field
    values in the Triangle."""

    sizes = set()
    for cell in triangle:
        for field in cell.values.values():
            if isinstance(field, np.ndarray):
                sizes |= {field.size}

    return sorted(sizes - {1})


def array_size(triangle: Triangle) -> int:
    """Get the consistent non-scalar array size for field values in the Triangle.
    If there is more than one distinct array size, raise an error."""

    sizes = array_sizes(triangle)
    if not sizes:
        return 1
    elif len(sizes) == 1:
        return sizes[0]
    else:
        raise ValueError(f"Multiple array sizes found in triangle: {sizes}")
